@@ -1,7 +1,7 @@
 /*
  *  libfaketime wrapper command
  *
- *  This file is part of libfaketime, version 0.9.6
+ *  This file is part of libfaketime, version 0.9.7
  *
  *  libfaketime is free software; you can redistribute it and/or modify it
  *  under the terms of the GNU General Public License v2 as published by the
@@ -48,7 +48,7 @@
 
 #include "faketime_common.h"
 
-const char version[] = "0.9.6";
+const char version[] = "0.9.7";
 
 #ifdef __APPLE__
 static const char *date_cmd = "gdate";
@@ -63,28 +63,29 @@ char sem_name[PATH_BUFSIZE] = {0}, shm_name[PATH_BUFSIZE] = {0};
 
 void usage(const char *name)
 {
-  printf("\n");
-  printf("Usage: %s [switches] <timestamp> <program with arguments>\n", name);
-  printf("\n");
-  printf("This will run the specified 'program' with the given 'arguments'.\n");
-  printf("The program will be tricked into seeing the given 'timestamp' as its starting date and time.\n");
-  printf("The clock will continue to run from this timestamp. Please see the manpage (man faketime)\n");
-  printf("for advanced options, such as stopping the wall clock and make it run faster or slower.\n");
-  printf("\n");
-  printf("The optional switches are:\n");
-  printf("  -m        : Use the multi-threaded version of libfaketime\n");
-  printf("  -f        : Use the advanced timestamp specification format (see manpage)\n");
-  printf("\n");
-  printf("Examples:\n");
-  printf("%s 'last friday 5 pm' /bin/date\n", name);
-  printf("%s '2008-12-24 08:15:42' /bin/date\n", name);
-  printf("%s -f '+2,5y x10,0' /bin/bash -c 'date; while true; do echo $SECONDS ; sleep 1 ; done'\n", name);
-  printf("%s -f '+2,5y x0,50' /bin/bash -c 'date; while true; do echo $SECONDS ; sleep 1 ; done'\n", name);
-  printf("%s -f '+2,5y i2,0' /bin/bash -c 'date; while true; do date; sleep 1 ; done'\n", name);
-  printf("In this single case all spawned processes will use the same global clock\n");
-  printf("without restaring it at the start of each process.\n\n");
-  printf("(Please note that it depends on your locale settings whether . or , has to be used for fractions)\n");
-  printf("\n");
+  printf("\n"
+  "Usage: %s [switches] <timestamp> <program with arguments>\n"
+  "\n"
+  "This will run the specified 'program' with the given 'arguments'.\n"
+  "The program will be tricked into seeing the given 'timestamp' as its starting date and time.\n"
+  "The clock will continue to run from this timestamp. Please see the manpage (man faketime)\n"
+  "for advanced options, such as stopping the wall clock and make it run faster or slower.\n"
+  "\n"
+  "The optional switches are:\n"
+  "  -m                  : Use the multi-threaded version of libfaketime\n"
+  "  -f                  : Use the advanced timestamp specification format (see manpage)\n"
+  "  --exclude-monotonic : Prevent monotonic clock from drifting (not the raw monotonic one)\n"
+  "\n"
+  "Examples:\n"
+  "%s 'last friday 5 pm' /bin/date\n"
+  "%s '2008-12-24 08:15:42' /bin/date\n"
+  "%s -f '+2,5y x10,0' /bin/bash -c 'date; while true; do echo $SECONDS ; sleep 1 ; done'\n"
+  "%s -f '+2,5y x0,50' /bin/bash -c 'date; while true; do echo $SECONDS ; sleep 1 ; done'\n"
+  "%s -f '+2,5y i2,0' /bin/bash -c 'date; while true; do date; sleep 1 ; done'\n"
+  "In this single case all spawned processes will use the same global clock\n"
+  "without restarting it at the start of each process.\n\n"
+  "(Please note that it depends on your locale settings whether . or , has to be used for fractions)\n"
+  "\n", name, name, name, name, name, name);
 }
 
 /** Clean up shared objects */
@@ -105,7 +106,6 @@ int main (int argc, char **argv)
   pid_t child_pid;
   int curr_opt = 1;
   bool use_mt = false, use_direct = false;
-  int pfds[2];
   long offset;
 
   while(curr_opt < argc)
@@ -122,6 +122,12 @@ int main (int argc, char **argv)
       curr_opt++;
       continue;
     }
+    else if (0 == strcmp(argv[curr_opt], "--exclude-monotonic"))
+    {
+      setenv("DONT_FAKE_MONOTONIC", "1", true);
+      curr_opt++;
+      continue;
+    }
     else if ((0 == strcmp(argv[curr_opt], "-v")) ||
              (0 == strcmp(argv[curr_opt], "--version")))
     {
@@ -154,6 +160,7 @@ int main (int argc, char **argv)
   if (!use_direct)
   {
     // TODO get seconds
+    int pfds[2];
     (void) (pipe(pfds) + 1);
     int ret = EXIT_SUCCESS;
 
@@ -183,6 +190,7 @@ int main (int argc, char **argv)
       offset = atol(buf) - time(NULL);
       ret = snprintf(buf, sizeof(buf), "%s%ld", (offset >= 0)?"+":"", offset);
       setenv("FAKETIME", buf, true);
+      close(pfds[0]); /* finished reading */
     }
   }
   else
@@ -190,6 +198,8 @@ int main (int argc, char **argv)
     /* simply pass format string along */
     setenv("FAKETIME", argv[curr_opt], true);
   }
+  int keepalive_fds[2];
+  (void) (pipe(keepalive_fds) + 1);
 
   /* we just consumed the timestamp option */
   curr_opt++;
@@ -201,8 +211,15 @@ int main (int argc, char **argv)
     struct ft_shared_s *ft_shared;
     char shared_objs[PATH_BUFSIZE];
 
-    snprintf(sem_name, PATH_BUFSIZE -1 ,"/faketime_sem_%d", getpid());
-    snprintf(shm_name, PATH_BUFSIZE -1 ,"/faketime_shm_%d", getpid());
+    /*
+     * Casting of getpid() return value to long needed to make GCC on SmartOS
+     * happy, since getpid's return value's type on SmartOS is long. Since
+     * getpid's return value's type is int on most other systems, and that
+     * sizeof(long) always >= sizeof(int), this works on all platforms without
+     * the need for crazy #ifdefs.
+     */
+    snprintf(sem_name, PATH_BUFSIZE -1 ,"/faketime_sem_%ld", (long)getpid());
+    snprintf(shm_name, PATH_BUFSIZE -1 ,"/faketime_shm_%ld", (long)getpid());
 
     if (SEM_FAILED == (sem = sem_open(sem_name, O_CREAT|O_EXCL, S_IWUSR|S_IRUSR, 1)))
     {
@@ -326,6 +343,7 @@ int main (int argc, char **argv)
   /* run command and clean up shared objects */
   if (0 == (child_pid = fork()))
   {
+    close(keepalive_fds[0]); /* only parent needs to read this */
     if (EXIT_SUCCESS != execvp(argv[curr_opt], &argv[curr_opt]))
     {
       perror("Running specified command failed");
@@ -335,7 +353,10 @@ int main (int argc, char **argv)
   else
   {
     int ret;
+    char buf;
+    close(keepalive_fds[1]); /* only children need keep this open */
     waitpid(child_pid, &ret, 0);
+    (void) (read(keepalive_fds[0], &buf, 1) + 1); /* reads 0B when all children exit */
     cleanup_shobjs();
     if (WIFSIGNALED(ret))
     {
@@ -344,6 +365,8 @@ int main (int argc, char **argv)
     }
     exit(WEXITSTATUS(ret));
   }
+
+  return EXIT_SUCCESS;
 }
 
 /*
