@@ -1,5 +1,5 @@
 /*
- *  This file is part of libfaketime, version 0.9.6
+ *  This file is part of libfaketime, version 0.9.7
  *
  *  libfaketime is free software; you can redistribute it and/or modify it
  *  under the terms of the GNU General Public License v2 as published by the
@@ -22,7 +22,6 @@
  */
 
 #define _GNU_SOURCE             /* required to get RTLD_NEXT defined */
-#define _XOPEN_SOURCE           /* required to get strptime() defined */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -57,12 +56,28 @@
 
 #ifndef __APPLE__
 extern char *__progname;
+#ifdef __sun
+#include "sunos_endian.h"
+#else
 #include <endian.h>
+#endif
 #else
 /* endianness related macros */
+#ifndef OSSwapHostToBigInt64
+#define OSSwapHostToBigInt64(x) ((uint64_t)(x))
+#endif
 #define htobe64(x) OSSwapHostToBigInt64(x)
+#ifndef OSSwapHostToLittleInt64
+#define OSSwapHostToLittleInt64(x) OSSwapInt64(x)
+#endif
 #define htole64(x) OSSwapHostToLittleInt64(x)
+#ifndef OSSwapBigToHostInt64
+#define OSSwapBigToHostInt64(x) ((uint64_t)(x))
+#endif
 #define be64toh(x) OSSwapBigToHostInt64(x)
+#ifndef OSSwapLittleToHostInt64
+#define OSSwapLittleToHostInt64(x) OSSwapInt64(x)
+#endif
 #define le64toh(x) OSSwapLittleToHostInt64(x)
 
 /* clock_gettime() and related clock definitions are missing on __APPLE__ */
@@ -93,7 +108,7 @@ typedef int clockid_t;
  * Per thread variable, which we turn on inside real_* calls to avoid modifying
  * time multiple times of for the whole process to prevent faking time
  */
-__thread bool dont_fake = false;
+static __thread bool dont_fake = false;
 
 /* Wrapper for function calls, which we want to return system time */
 #define DONT_FAKE_TIME(call)          \
@@ -125,7 +140,7 @@ static int          (*real___ftime)           (struct timeb *);
 static int          (*real___gettimeofday)    (struct timeval *, void *);
 static int          (*real___clock_gettime)   (clockid_t clk_id, struct timespec *tp);
 #endif
-#ifndef __APPLE__
+#ifndef __APPLEOSX__
 #ifdef FAKE_TIMERS
 static int          (*real_timer_settime_22)   (int timerid, int flags, const struct itimerspec *new_value,
                                                 struct itimerspec * old_value);
@@ -145,17 +160,23 @@ static unsigned int (*real_sleep)           (unsigned int seconds);
 static unsigned int (*real_alarm)           (unsigned int seconds);
 static int          (*real_poll)            (struct pollfd *, nfds_t, int);
 static int          (*real_ppoll)           (struct pollfd *, nfds_t, const struct timespec *, const sigset_t *);
+static int          (*real_select)          (int nfds, fd_set *restrict readfds,
+                                             fd_set *restrict writefds,
+                                             fd_set *restrict errorfds,
+                                             struct timeval *restrict timeout);
+static int          (*real_sem_timedwait)   (sem_t*, const struct timespec*);
 #endif
-#ifdef __APPLE__
+#ifdef __APPLEOSX__
 static int          (*real_clock_get_time)  (clock_serv_t clock_serv, mach_timespec_t *cur_timeclockid_t);
 static int          apple_clock_gettime     (clockid_t clk_id, struct timespec *tp);
 static clock_serv_t clock_serv_real;
 #endif
 
+static int initialized = 0;
+
 /* prototypes */
-time_t fake_time(time_t *time_tptr);
-int    fake_gettimeofday(struct timeval *tv);
-int    fake_clock_gettime(clockid_t clk_id, struct timespec *tp);
+static int    fake_gettimeofday(struct timeval *tv);
+static int    fake_clock_gettime(clockid_t clk_id, struct timespec *tp);
 
 /** Semaphore protecting shared data */
 static sem_t *shared_sem = NULL;
@@ -199,11 +220,19 @@ static char ft_spawn_target[1024];
 static long ft_spawn_secs = -1;
 static long ft_spawn_ncalls = -1;
 
+static int fake_monotonic_clock = 1;
+static int cache_enabled = 1;
+static int cache_duration = 10;     /* cache fake time input for 10 seconds */
+
 /*
  * Static timespec to store our startup time, followed by a load-time library
  * initialization declaration.
  */
+#ifndef CLOCK_BOOTTIME
 static struct system_time_s ftpl_starttime = {{0, -1}, {0, -1}, {0, -1}};
+#else
+static struct system_time_s ftpl_starttime = {{0, -1}, {0, -1}, {0, -1}, {0, -1}};
+#endif
 
 static char user_faked_time_fmt[BUFSIZ] = {0};
 
@@ -211,6 +240,8 @@ static char user_faked_time_fmt[BUFSIZ] = {0};
 static struct timespec user_faked_time_timespec = {0, -1};
 /* User supplied base time is set */
 static bool user_faked_time_set = false;
+static char user_faked_time_saved[BUFFERLEN] = {0};
+
 /* Fractional user offset provided through FAKETIME env. var.*/
 static struct timespec user_offset = {0, -1};
 /* Speed up or slow down clock */
@@ -225,6 +256,7 @@ enum ft_mode_t {FT_FREEZE, FT_START_AT, FT_NOOP} ft_mode = FT_FREEZE;
 static bool parse_config_file = true;
 
 void ft_cleanup (void) __attribute__ ((destructor));
+void ftpl_init (void) __attribute__ ((constructor));
 
 
 /*
@@ -237,6 +269,7 @@ static void ft_shm_init (void)
 {
   int ticks_shm_fd;
   char sem_name[256], shm_name[256], *ft_shared_env = getenv("FAKETIME_SHARED");
+
   if (ft_shared_env != NULL)
   {
     if (sscanf(ft_shared_env, "%255s %255s", sem_name, shm_name) < 2)
@@ -293,7 +326,7 @@ void ft_cleanup (void)
 /* Get system time from system for all clocks */
 static void system_time_from_system (struct system_time_s * systime)
 {
-#ifdef __APPLE__
+#ifdef __APPLEOSX__
   /* from http://stackoverflow.com/questions/5167269/clock-gettime-alternative-in-mac-os-x */
   clock_serv_t cclock;
   mach_timespec_t mts;
@@ -309,9 +342,16 @@ static void system_time_from_system (struct system_time_s * systime)
   systime->mon_raw.tv_sec = mts.tv_sec;
   systime->mon_raw.tv_nsec = mts.tv_nsec;
 #else
-  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_REALTIME, &systime->real));
-  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_MONOTONIC, &systime->mon));
-  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_MONOTONIC_RAW, &systime->mon_raw));
+  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_REALTIME, &systime->real))
+   ;
+  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_MONOTONIC, &systime->mon))
+   ;
+  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_MONOTONIC_RAW, &systime->mon_raw))
+   ;
+#ifdef CLOCK_BOOTTIME
+  DONT_FAKE_TIME((*real_clock_gettime)(CLOCK_BOOTTIME, &systime->boot))
+   ;
+#endif
 #endif
 }
 
@@ -495,6 +535,10 @@ static inline void fake_stat64buf (struct stat64 *buf) {
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __xstat (int ver, const char *path, struct stat *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_stat)
   { /* dlsym() failed */
 #ifdef DEBUG
@@ -524,6 +568,10 @@ int __xstat (int ver, const char *path, struct stat *buf)
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __fxstat (int ver, int fildes, struct stat *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_fstat)
   {  /* dlsym() failed */
 #ifdef DEBUG
@@ -553,6 +601,10 @@ int __fxstat (int ver, int fildes, struct stat *buf)
 #ifndef NO_ATFILE
 int __fxstatat(int ver, int fildes, const char *filename, struct stat *buf, int flag)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_fstatat)
   { /* dlsym() failed */
 #ifdef DEBUG
@@ -582,6 +634,10 @@ int __fxstatat(int ver, int fildes, const char *filename, struct stat *buf, int
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __lxstat (int ver, const char *path, struct stat *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_lstat)
   {  /* dlsym() failed */
 #ifdef DEBUG
@@ -610,6 +666,10 @@ int __lxstat (int ver, const char *path, struct stat *buf)
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __xstat64 (int ver, const char *path, struct stat64 *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_stat64)
   { /* dlsym() failed */
 #ifdef DEBUG
@@ -638,6 +698,10 @@ int __xstat64 (int ver, const char *path, struct stat64 *buf)
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __fxstat64 (int ver, int fildes, struct stat64 *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_fstat64)
   {  /* dlsym() failed */
 #ifdef DEBUG
@@ -667,6 +731,10 @@ int __fxstat64 (int ver, int fildes, struct stat64 *buf)
 #ifndef NO_ATFILE
 int __fxstatat64 (int ver, int fildes, const char *filename, struct stat64 *buf, int flag)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_fstatat64)
   {  /* dlsym() failed */
 #ifdef DEBUG
@@ -696,6 +764,10 @@ int __fxstatat64 (int ver, int fildes, const char *filename, struct stat64 *buf,
 /* Contributed by Philipp Hachtmann in version 0.6 */
 int __lxstat64 (int ver, const char *path, struct stat64 *buf)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (NULL == real_lstat64)
   {  /* dlsym() failed */
 #ifdef DEBUG
@@ -738,6 +810,10 @@ int nanosleep(const struct timespec *req, struct timespec *rem)
   int result;
   struct timespec real_req;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_nanosleep == NULL)
   {
     return -1;
@@ -782,6 +858,11 @@ int nanosleep(const struct timespec *req, struct timespec *rem)
 int usleep(useconds_t usec)
 {
   int result;
+
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (user_rate_set && !dont_fake)
   {
     struct timespec real_req;
@@ -814,6 +895,10 @@ int usleep(useconds_t usec)
  */
 unsigned int sleep(unsigned int seconds)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (user_rate_set && !dont_fake)
   {
     if (real_nanosleep == NULL)
@@ -865,6 +950,11 @@ unsigned int alarm(unsigned int seconds)
 {
   unsigned int ret;
   unsigned int seconds_real = (user_rate_set && !dont_fake)?((1.0 / user_rate) * seconds):seconds;
+
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_alarm == NULL)
   {
     return -1;
@@ -883,6 +973,10 @@ int ppoll(struct pollfd *fds, nfds_t nfds,
   struct timespec real_timeout, *real_timeout_pt;
   int ret;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_ppoll == NULL)
   {
     return -1;
@@ -915,6 +1009,11 @@ int ppoll(struct pollfd *fds, nfds_t nfds,
 int poll(struct pollfd *fds, nfds_t nfds, int timeout)
 {
   int ret, timeout_real = (user_rate_set && !dont_fake && (timeout > 0))?(timeout / user_rate):timeout;
+
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_poll == NULL)
   {
     return -1;
@@ -923,6 +1022,98 @@ int poll(struct pollfd *fds, nfds_t nfds, int timeout)
   DONT_FAKE_TIME(ret = (*real_poll)(fds, nfds, timeout_real));
   return ret;
 }
+
+/*
+ * Faked select()
+ */
+int select(int nfds, fd_set *readfds,
+           fd_set *writefds,
+           fd_set *errorfds,
+           struct timeval *timeout)
+{
+  int ret;
+  struct timeval timeout_real;
+
+  if (!initialized)
+  {
+    ftpl_init();
+  }
+
+  if (real_select == NULL)
+  {
+    return -1;
+  }
+
+  if (timeout != NULL)
+  {
+    if (user_rate_set && !dont_fake && (timeout->tv_sec > 0 || timeout->tv_usec > 0))
+    {
+      struct timespec ts;
+
+      ts.tv_sec = timeout->tv_sec;
+      ts.tv_nsec = timeout->tv_usec * 1000;
+
+      timespecmul(&ts, 1.0 / user_rate, &ts);
+
+      timeout_real.tv_sec = ts.tv_sec;
+      timeout_real.tv_usec = ts.tv_nsec / 1000;
+    }
+    else
+    {
+      timeout_real.tv_sec = timeout->tv_sec;
+      timeout_real.tv_usec = timeout->tv_usec;
+    }
+  }
+
+  DONT_FAKE_TIME(ret = (*real_select)(nfds, readfds, writefds, errorfds, timeout == NULL ? timeout : &timeout_real));
+  return ret;
+}
+
+int sem_timedwait(sem_t *sem, const struct timespec *abs_timeout)
+{
+  int result;
+  struct timespec real_abs_timeout, *real_abs_timeout_pt;
+
+  /* sanity check */
+  if (abs_timeout == NULL)
+  {
+    return -1;
+  }
+
+  if (NULL == real_sem_timedwait)
+  {  /* dlsym() failed */
+#ifdef DEBUG
+    (void) fprintf(stderr, "faketime problem: original sem_timedwait() not found.\n");
+#endif
+    return -1; /* propagate error to caller */
+  }
+
+  if (!dont_fake)
+  {
+    struct timespec tdiff, timeadj;
+
+    timespecsub(abs_timeout, &ftpl_starttime.real, &tdiff);
+
+    if (user_rate_set)
+    {
+      timespecmul(&tdiff, user_rate, &timeadj);
+    }
+    else
+    {
+        timeadj = tdiff;
+    }
+    timespecadd(&user_faked_time_timespec, &timeadj, &real_abs_timeout);
+    real_abs_timeout_pt = &real_abs_timeout;
+  }
+  else
+  {
+    /* cast away constness */
+    real_abs_timeout_pt = (struct timespec *)abs_timeout;
+  }
+
+  DONT_FAKE_TIME(result = (*real_sem_timedwait)(sem, real_abs_timeout_pt));
+  return result;
+}
 #endif
 
 #ifndef __APPLE__
@@ -953,6 +1144,10 @@ timer_settime_common(timer_t_or_int timerid, int flags,
   struct itimerspec new_real;
   struct itimerspec *new_real_pt = &new_real;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (new_value == NULL)
   {
     new_real_pt = NULL;
@@ -1057,6 +1252,10 @@ int timer_settime_22(int timerid, int flags,
          const struct itimerspec *new_value,
          struct itimerspec *old_value)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_timer_settime_22 == NULL)
   {
     return -1;
@@ -1075,6 +1274,10 @@ int timer_settime_233(timer_t timerid, int flags,
       const struct itimerspec *new_value,
       struct itimerspec *old_value)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_timer_settime_233 == NULL)
   {
     return -1;
@@ -1094,6 +1297,10 @@ int timer_gettime_common(timer_t_or_int timerid, struct itimerspec *curr_value,
 {
   int result;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_timer_gettime_233 == NULL)
   {
     return -1;
@@ -1135,6 +1342,10 @@ int timer_gettime_common(timer_t_or_int timerid, struct itimerspec *curr_value,
  */
 int timer_gettime_22(timer_t timerid, struct itimerspec *curr_value)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_timer_gettime_22 == NULL)
   {
     return -1;
@@ -1151,6 +1362,10 @@ int timer_gettime_22(timer_t timerid, struct itimerspec *curr_value)
  */
 int timer_gettime_233(timer_t timerid, struct itimerspec *curr_value)
 {
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   if (real_timer_gettime_233 == NULL)
   {
     return -1;
@@ -1186,6 +1401,10 @@ time_t time(time_t *time_tptr)
   struct timespec tp;
   time_t result;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   DONT_FAKE_TIME(result = (*real_clock_gettime)(CLOCK_REALTIME, &tp));
   if (result == -1) return -1;
 
@@ -1204,6 +1423,10 @@ int ftime(struct timeb *tb)
   struct timespec tp;
   int result;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   /* sanity check */
   if (tb == NULL)
     return 0;               /* ftime() always returns 0, see manpage */
@@ -1241,6 +1464,10 @@ int gettimeofday(struct timeval *tv, void *tz)
 {
   int result;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   /* sanity check */
   if (tv == NULL)
   {
@@ -1271,6 +1498,10 @@ int clock_gettime(clockid_t clk_id, struct timespec *tp)
 {
   int result;
 
+  if (!initialized)
+  {
+    ftpl_init();
+  }
   /* sanity check */
   if (tp == NULL)
   {
@@ -1290,7 +1521,10 @@ int clock_gettime(clockid_t clk_id, struct timespec *tp)
   if (result == -1) return result; /* original function failed */
 
   /* pass the real current time to our faking version, overwriting it */
-  result = fake_clock_gettime(clk_id, tp);
+  if (fake_monotonic_clock || clk_id != CLOCK_MONOTONIC)
+  {
+    result = fake_clock_gettime(clk_id, tp);
+  }
 
   /* return the result to the caller */
   return result;
@@ -1307,6 +1541,13 @@ static void parse_ft_string(const char *user_faked_time)
 {
   struct tm user_faked_time_tm;
   char * tmp_time_fmt;
+
+  if (!strncmp(user_faked_time, user_faked_time_saved, BUFFERLEN))
+  {
+      /* No change */
+      return;
+  }
+
   /* check whether the user gave us an absolute time to fake */
   switch (user_faked_time[0])
   {
@@ -1355,6 +1596,9 @@ static void parse_ft_string(const char *user_faked_time)
 
       user_faked_time_timespec.tv_sec = mktime(&user_faked_time_tm);
       user_faked_time_timespec.tv_nsec = 0;
+
+      /* Reset starttime */
+      system_time_from_system(&ftpl_starttime);
       goto parse_modifiers;
       break;
 
@@ -1377,6 +1621,12 @@ parse_modifiers:
       }
       break;
   } // end of switch
+
+  strncpy(user_faked_time_saved, user_faked_time, BUFFERLEN-1);
+  user_faked_time_saved[BUFFERLEN-1] = 0;
+#ifdef DEBUG
+  fprintf(stderr, "new FAKETIME: %s\n", user_faked_time_saved);
+#endif
 }
 
 
@@ -1386,9 +1636,10 @@ parse_modifiers:
  *      =======================================================================
  */
 
-void __attribute__ ((constructor)) ftpl_init(void)
+void ftpl_init(void)
 {
   char *tmp_env;
+  bool dont_fake_final;
 
 #ifdef __APPLE__
   const char *progname = getprogname();
@@ -1415,18 +1666,28 @@ void __attribute__ ((constructor)) ftpl_init(void)
   real_alarm =              dlsym(RTLD_NEXT, "alarm");
   real_poll =               dlsym(RTLD_NEXT, "poll");
   real_ppoll =              dlsym(RTLD_NEXT, "ppoll");
+  real_select =             dlsym(RTLD_NEXT, "select");
+  real_sem_timedwait =      dlsym(RTLD_NEXT, "sem_timedwait");
 #endif
 #ifdef FAKE_INTERNAL_CALLS
   real___ftime =              dlsym(RTLD_NEXT, "__ftime");
   real___gettimeofday =       dlsym(RTLD_NEXT, "__gettimeofday");
   real___clock_gettime  =     dlsym(RTLD_NEXT, "__clock_gettime");
 #endif
-#ifdef __APPLE__
+#ifdef __APPLEOSX__
   real_clock_get_time =     dlsym(RTLD_NEXT, "clock_get_time");
   real_clock_gettime  =     apple_clock_gettime;
 #else
-  real_clock_gettime  =     dlsym(RTLD_NEXT, "clock_gettime");
+  real_clock_gettime  =     dlsym(RTLD_NEXT, "__clock_gettime");
+  if (NULL == real_clock_gettime)
+  {
+    real_clock_gettime  =   dlsym(RTLD_NEXT, "clock_gettime");
+  }
 #ifdef FAKE_TIMERS
+#if defined(__sun)
+    real_timer_gettime_233 =  dlsym(RTLD_NEXT, "timer_gettime");
+    real_timer_settime_233 =  dlsym(RTLD_NEXT, "timer_settime");
+#else
   real_timer_settime_22 =   dlvsym(RTLD_NEXT, "timer_settime","GLIBC_2.2");
   real_timer_settime_233 =  dlvsym(RTLD_NEXT, "timer_settime","GLIBC_2.3.3");
   if (NULL == real_timer_settime_233)
@@ -1441,6 +1702,11 @@ void __attribute__ ((constructor)) ftpl_init(void)
   }
 #endif
 #endif
+#endif
+
+  dont_fake = true; // Do not fake times during initialization
+  dont_fake_final = false;
+  initialized = 1;
 
   ft_shm_init();
 #ifdef FAKE_STAT
@@ -1450,29 +1716,59 @@ void __attribute__ ((constructor)) ftpl_init(void)
   }
 #endif
 
+  if ((tmp_env = getenv("FAKETIME_CACHE_DURATION")) != NULL)
+  {
+    cache_duration = atoi(tmp_env);
+  }
+  if ((tmp_env = getenv("FAKETIME_NO_CACHE")) != NULL)
+  {
+    if (0 == strcmp(tmp_env, "1"))
+    {
+      cache_enabled = 0;
+    }
+  }
+  if ((tmp_env = getenv("DONT_FAKE_MONOTONIC")) != NULL)
+  {
+    if (0 == strcmp(tmp_env, "1"))
+    {
+      fake_monotonic_clock = 0;
+    }
+  }
   /* Check whether we actually should be faking the returned timestamp. */
 
   /* We can prevent faking time for specified commands */
   if ((tmp_env = getenv("FAKETIME_SKIP_CMDS")) != NULL)
   {
-    char *skip_cmd, *saveptr;
-    skip_cmd = strtok_r(tmp_env, ",", &saveptr);
-    while (skip_cmd != NULL)
+    char *skip_cmd, *saveptr, *tmpvar;
+    /* Don't mess with the env variable directly. */
+    tmpvar = strdup(tmp_env);
+    if (tmpvar != NULL)
     {
-      if (0 == strcmp(progname, skip_cmd))
+      skip_cmd = strtok_r(tmpvar, ",", &saveptr);
+      while (skip_cmd != NULL)
       {
-        ft_mode = FT_NOOP;
-        dont_fake = true;
-        break;
+        if (0 == strcmp(progname, skip_cmd))
+        {
+          ft_mode = FT_NOOP;
+          dont_fake_final = true;
+          break;
+        }
+        skip_cmd = strtok_r(NULL, ",", &saveptr);
       }
-      skip_cmd = strtok_r(NULL, ",", &saveptr);
+      free(tmpvar);
+      tmpvar = NULL;
+    }
+    else
+    {
+      fprintf(stderr, "Error: Could not copy the environment variable value.\n");
+      exit(EXIT_FAILURE);
     }
   }
 
   /* We can limit faking time to specified commands */
   if ((tmp_env = getenv("FAKETIME_ONLY_CMDS")) != NULL)
   {
-    char *only_cmd, *saveptr;
+    char *only_cmd, *saveptr, *tmpvar;
     bool cmd_matched = false;
 
     if (getenv("FAKETIME_SKIP_CMDS") != NULL)
@@ -1481,21 +1777,29 @@ void __attribute__ ((constructor)) ftpl_init(void)
       exit(EXIT_FAILURE);
     }
 
-    only_cmd = strtok_r(tmp_env, ",", &saveptr);
-    while (only_cmd != NULL)
-    {
-      if (0 == strcmp(progname, only_cmd))
+    /* Don't mess with the env variable directly. */
+    tmpvar = strdup(tmp_env);
+    if (tmpvar != NULL) {
+      only_cmd = strtok_r(tmpvar, ",", &saveptr);
+      while (only_cmd != NULL)
       {
-        cmd_matched = true;
-        break;
+        if (0 == strcmp(progname, only_cmd))
+        {
+          cmd_matched = true;
+          break;
+        }
+        only_cmd = strtok_r(NULL, ",", &saveptr);
       }
-      only_cmd = strtok_r(NULL, ",", &saveptr);
-    }
 
-    if (!cmd_matched)
-    {
-      ft_mode = FT_NOOP;
-      dont_fake = true;
+      if (!cmd_matched)
+      {
+        ft_mode = FT_NOOP;
+        dont_fake_final = true;
+      }
+      free(tmpvar);
+    } else {
+      fprintf(stderr, "Error: Could not copy the environment variable value.\n");
+      exit(EXIT_FAILURE);
     }
   }
 
@@ -1622,6 +1926,8 @@ void __attribute__ ((constructor)) ftpl_init(void)
     parse_config_file = false;
     parse_ft_string(tmp_env);
   }
+
+  dont_fake = dont_fake_final;
 }
 
 
@@ -1659,7 +1965,6 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
   /* variables used for caching, introduced in version 0.6 */
   static time_t last_data_fetch = 0;  /* not fetched previously at first call */
   static int cache_expired = 1;       /* considered expired at first call */
-  static int cache_duration = 10;     /* cache fake time input for 10 seconds */
 
   if (dont_fake) return 0;
   /* Per process timers are only sped up or slowed down */
@@ -1695,14 +2000,25 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
     switch (clk_id)
     {
       case CLOCK_REALTIME:
+#ifdef CLOCK_REALTIME_COARSE
+      case CLOCK_REALTIME_COARSE:
+#endif
         timespecsub(tp, &ftpl_starttime.real, &tmp_ts);
         break;
       case CLOCK_MONOTONIC:
+#ifdef CLOCK_MONOTONIC_COARSE
+      case CLOCK_MONOTONIC_COARSE:
+#endif
         timespecsub(tp, &ftpl_starttime.mon, &tmp_ts);
         break;
       case CLOCK_MONOTONIC_RAW:
         timespecsub(tp, &ftpl_starttime.mon_raw, &tmp_ts);
         break;
+#ifdef CLOCK_BOOTTIME
+      case CLOCK_BOOTTIME:
+        timespecsub(tp, &ftpl_starttime.boot, &tmp_ts);
+        break;
+#endif
       default:
         printf("Invalid clock_id for clock_gettime: %d", clk_id);
         exit(EXIT_FAILURE);
@@ -1745,40 +2061,45 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
     }
   }
 
-#ifdef NO_CACHING
-  cache_expired = 1;
-#endif
+  if (cache_enabled == 0)
+  {
+    cache_expired = 1;
+  }
 
   if (cache_expired == 1)
   {
-    last_data_fetch = tp->tv_sec;
+    static char user_faked_time[BUFFERLEN]; /* changed to static for caching in v0.6 */
+    /* initialize with default or env. variable */
+    char *tmp_env;
+
     /* Can be enabled for testing ...
       fprintf(stderr, "***************++ Cache expired ++**************\n");
     */
 
+    if (NULL != (tmp_env = getenv("FAKETIME")))
+    {
+      strncpy(user_faked_time, tmp_env, BUFFERLEN);
+    }
+    else
+    {
+      snprintf(user_faked_time, BUFFERLEN, "+0");
+    }
+
+    last_data_fetch = tp->tv_sec;
     /* fake time supplied as environment variable? */
     if (parse_config_file)
     {
-      static char user_faked_time[BUFFERLEN]; /* changed to static for caching in v0.6 */
+      char custom_filename[BUFSIZ];
       char filename[BUFSIZ];
       FILE *faketimerc;
-      /* initialize with default or env. variable */
-      char *tmp_env;
-      if (NULL != (tmp_env = getenv("FAKETIME")))
-      {
-        strncpy(user_faked_time, tmp_env, BUFFERLEN);
-      }
-      else
-      {
-        snprintf(user_faked_time, BUFFERLEN, "+0");
-      }
-
       /* check whether there's a .faketimerc in the user's home directory, or
        * a system-wide /etc/faketimerc present.
        * The /etc/faketimerc handling has been contributed by David Burley,
        * Jacob Moorman, and Wayne Davison of SourceForge, Inc. in version 0.6 */
+      (void) snprintf(custom_filename, BUFSIZ, "%s", getenv("FAKETIME_TIMESTAMP_FILE"));
       (void) snprintf(filename, BUFSIZ, "%s/.faketimerc", getenv("HOME"));
-      if ((faketimerc = fopen(filename, "rt")) != NULL ||
+      if ((faketimerc = fopen(custom_filename, "rt")) != NULL ||
+          (faketimerc = fopen(filename, "rt")) != NULL ||
           (faketimerc = fopen("/etc/faketimerc", "rt")) != NULL)
       {
         char line[BUFFERLEN];
@@ -1795,8 +2116,8 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
         }
         fclose(faketimerc);
       }
-      parse_ft_string(user_faked_time);
     } /* read fake time from file */
+    parse_ft_string(user_faked_time);
   } /* cache had expired */
 
   if (infile_set)
@@ -1830,14 +2151,25 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
         switch (clk_id)
         {
           case CLOCK_REALTIME:
+#ifdef CLOCK_REALTIME_COARSE
+          case CLOCK_REALTIME_COARSE:
+#endif
             timespecsub(tp, &ftpl_starttime.real, &tdiff);
             break;
           case CLOCK_MONOTONIC:
+#ifdef CLOCK_MONOTONIC_COARSE
+          case CLOCK_MONOTONIC_COARSE:
+#endif
             timespecsub(tp, &ftpl_starttime.mon, &tdiff);
             break;
           case CLOCK_MONOTONIC_RAW:
             timespecsub(tp, &ftpl_starttime.mon_raw, &tdiff);
             break;
+#ifdef CLOCK_BOOTTIME
+          case CLOCK_BOOTTIME:
+            timespecsub(tp, &ftpl_starttime.boot, &tdiff);
+            break;
+#endif
           default:
             printf("Invalid clock_id for clock_gettime: %d", clk_id);
             exit(EXIT_FAILURE);
@@ -1865,17 +2197,6 @@ int fake_clock_gettime(clockid_t clk_id, struct timespec *tp)
   return 0;
 }
 
-time_t fake_time(time_t *time_tptr)
-{
-  struct timespec tp;
-
-  tp.tv_sec = *time_tptr;
-  tp.tv_nsec = ftpl_starttime.real.tv_nsec;
-  (void)fake_clock_gettime(CLOCK_REALTIME, &tp);
-  *time_tptr = tp.tv_sec;
-  return *time_tptr;
-}
-
 int fake_gettimeofday(struct timeval *tv)
 {
   struct timespec ts;
@@ -1897,7 +2218,7 @@ int fake_gettimeofday(struct timeval *tv)
  *      =======================================================================
  */
 
-#ifdef __APPLE__
+#ifdef __APPLEOSX__
 /*
  * clock_gettime implementation for __APPLE__
  * @note It always behave like being called with CLOCK_REALTIME.
@@ -2006,7 +2327,10 @@ int __clock_gettime(clockid_t clk_id, struct timespec *tp)
   if (result == -1) return result; /* original function failed */
 
   /* pass the real current time to our faking version, overwriting it */
-  result = fake_clock_gettime(clk_id, tp);
+  if (fake_monotonic_clock || clk_id != CLOCK_MONOTONIC)
+  {
+    result = fake_clock_gettime(clk_id, tp);
+  }
 
   /* return the result to the caller */
   return result;
